/***************************************************************************
 *   Copyright (C) 2005 by Dominic Rath                                    *
 *   Dominic.Rath@gmx.de                                                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "config.h"

#include "xscale.h"

#include "register.h"
#include "target.h"
#include "armv4_5.h"
#include "log.h"
#include "jtag.h"
#include "binarybuffer.h"
#include "time_support.h"

#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

/* cli handling */
int xscale_register_commands(struct command_context_s *cmd_ctx);

/* forward declarations */
int xscale_target_command(struct command_context_s *cmd_ctx, char *cmd, char **args, int argc, struct target_s *target);
int xscale_init_target(struct command_context_s *cmd_ctx, struct target_s *target);
int xscale_quit();

enum target_state xscale_poll(target_t *target);
int xscale_halt(target_t *target);
int xscale_resume(struct target_s *target, int current, u32 address, int handle_breakpoints, int debug_execution);
int xscale_step(struct target_s *target, int current, u32 address, int handle_breakpoints);

int xscale_assert_reset(target_t *target);
int xscale_deassert_reset(target_t *target);
int xscale_soft_reset_halt(struct target_s *target);

int xscale_read_core_reg(struct target_s *target, int num, enum armv4_5_mode mode);
int xscale_write_core_reg(struct target_s *target, int num, enum armv4_5_mode mode, u32 value);

int xscale_read_memory(struct target_s *target, u32 address, u32 size, u32 count, u8 *buffer);
int xscale_write_memory(struct target_s *target, u32 address, u32 size, u32 count, u8 *buffer);
int xscale_bulk_write_memory(target_t *target, u32 address, u32 count, u8 *buffer);

int xscale_add_breakpoint(struct target_s *target, u32 address, u32 length, enum breakpoint_type type);
int xscale_remove_breakpoint(struct target_s *target, breakpoint_t *breakpoint);
int xscale_add_watchpoint(struct target_s *target, u32 address, u32 length, enum watchpoint_rw rw);
int xscale_remove_watchpoint(struct target_s *target, watchpoint_t *watchpoint);

target_type_t xscale_target =
{
	.name = "xscale",

	.poll = xscale_poll,
	.arch_state = armv4_5_arch_state,

	.halt = xscale_halt,
	.resume = xscale_resume,
	.step = xscale_step,

	.assert_reset = xscale_assert_reset,
	.deassert_reset = xscale_deassert_reset,
	.soft_reset_halt = xscale_soft_reset_halt,

	.get_gdb_reg_list = armv4_5_get_gdb_reg_list,
	
	.read_memory = xscale_read_memory,
	.write_memory = xscale_write_memory,
	.bulk_write_memory = xscale_bulk_write_memory,

	.run_algorithm = armv4_5_run_algorithm,
	
	.add_breakpoint = xscale_add_breakpoint,
	.remove_breakpoint = xscale_remove_breakpoint,
	.add_watchpoint = xscale_add_watchpoint,
	.remove_watchpoint = xscale_remove_watchpoint,

	.register_commands = xscale_register_commands,
	.target_command = xscale_target_command,
	.init_target = xscale_init_target,
	.quit = xscale_quit
};

char* xscale_reg_list[] =
{
	"XSCALE_MAINID",
	"XSCALE_CACHETYPE",
	"XSCALE_CTRL",
	"XSCALE_AUXCTRL",
	"XSCALE_TTB",
	"XSCALE_DAC",
	"XSCALE_FSR",
	"XSCALE_FAR",
	"XSCALE_PID",
	"XSCALE_CPACCESS",
	"XSCALE_IBCR0",
	"XSCALE_IBCR1",
	"XSCALE_DBR0",
	"XSCALE_DBR1",
	"XSCALE_DBCON",
	"XSCALE_TX",
	"XSCALE_RX",
	"XSCALE_TXRXCTRL",
	"XSCALE_DCSR",
	"XSCALE_TBREG",
	"XSCALE_CHKPT0",
	"XSCALE_CHKPT1",
};

xscale_reg_t xscale_reg_arch_info[] =
{
	{XSCALE_MAINID, NULL},
	{XSCALE_CACHETYPE, NULL},
	{XSCALE_CTRL, NULL},
	{XSCALE_AUXCTRL, NULL},
	{XSCALE_TTB, NULL},
	{XSCALE_DAC, NULL},
	{XSCALE_FSR, NULL},
	{XSCALE_FAR, NULL},
	{XSCALE_PID, NULL},
	{XSCALE_CPACCESS, NULL},
	{XSCALE_IBCR0, NULL},
	{XSCALE_IBCR1, NULL},
	{XSCALE_DBR0, NULL},
	{XSCALE_DBR1, NULL},
	{XSCALE_DBCON, NULL},
	{XSCALE_TBREG, NULL},
	{XSCALE_CHKPT0, NULL},
	{XSCALE_CHKPT1, NULL},
	{-1, NULL}, /* TX */
	{-1, NULL}, /* RX */
	{-1, NULL}, /* TXRXCTRL */
	{-1, NULL}, /* DCSR */
};

int xscale_reg_arch_type = -1;

int xscale_get_reg(reg_t *reg);
int xscale_set_reg(reg_t *reg, u32 value);

int xscale_jtag_set_instr(int chain_pos, u32 new_instr)
{
	jtag_device_t *device = jtag_get_device(chain_pos);
	
	if (buf_get_u32(device->cur_instr, 0, device->ir_length) != new_instr)
	{
		scan_field_t field;
	
		field.device = chain_pos;
		field.num_bits = device->ir_length;
		field.out_value = calloc(CEIL(field.num_bits, 8), 1);
		buf_set_u32(field.out_value, 0, field.num_bits, new_instr);
		field.out_mask = NULL;
		field.in_value = NULL;
		field.in_check_value = NULL;
		field.in_check_mask = NULL;
		field.in_handler = NULL;
		field.in_handler_priv = NULL;
		
		jtag_add_ir_scan(1, &field, -1);
		
		free(field.out_value);
	}
	
	return ERROR_OK;
}

int xscale_jtag_callback(enum jtag_event event, void *priv)
{
	target_t *target = priv;
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	switch (event)
	{
		case JTAG_TRST_ASSERTED:
			break;
		case JTAG_TRST_RELEASED:
			break;
		case JTAG_SRST_ASSERTED:
			break;
		case JTAG_SRST_RELEASED:
			break;
		default:
			WARNING("unhandled JTAG event");
	}
	
	return ERROR_OK;
}

int xscale_read_dcsr(target_t *target)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	int retval;
	
	scan_field_t fields[3];
	u8 field0 = 0x0;
	u8 field2 = 0x0;
	u8 verify = 0x0;

	jtag_add_end_state(TAP_PD);
	xscale_jtag_set_instr(xscale->chain_pos, 0x09); /* DCSR */
	
	buf_set_u32(&field0, 1, 1, xscale->hold_rst);
	buf_set_u32(&field0, 2, 1, xscale->external_debug_break);
	
	fields[0].device = xscale->chain_pos;
	fields[0].num_bits = 3;
	fields[0].out_value = &field0;
	fields[0].out_mask = NULL;
	fields[0].in_value = &verify;
	fields[0].in_check_value = NULL;
	fields[0].in_check_mask = NULL;
	fields[0].in_handler = NULL;
	fields[0].in_handler_priv = NULL;
		
	fields[1].device = xscale->chain_pos;
	fields[1].num_bits = 32;
	fields[1].out_value = NULL;
	fields[1].out_mask = NULL;
	fields[1].in_value = xscale->reg_cache->reg_list[XSCALE_DCSR].value;
	fields[1].in_handler = NULL;
	fields[1].in_handler_priv = NULL;
	fields[1].in_check_value = NULL;
	fields[1].in_check_mask = NULL;

	fields[2].device = xscale->chain_pos;
	fields[2].num_bits = 1;
	fields[2].out_value = &field2;
	fields[2].out_mask = NULL;
	fields[2].in_value = NULL;
	fields[2].in_check_value = NULL;
	fields[2].in_check_mask = NULL;
	fields[2].in_handler = NULL;
	fields[2].in_handler_priv = NULL;
	
	jtag_add_dr_scan(3, fields, -1);

	if ((retval = jtag_execute_queue()) != ERROR_OK)
	{
		ERROR("JTAG error while reading DCSR");
		exit(-1);
	}
	
	DEBUG("verify: 0x%x", verify);
	
	xscale->reg_cache->reg_list[XSCALE_DCSR].dirty = 0;
	xscale->reg_cache->reg_list[XSCALE_DCSR].valid = 1;
	
	fields[1].out_value = xscale->reg_cache->reg_list[XSCALE_DCSR].value;
	fields[1].in_value = NULL;
	
	jtag_add_end_state(TAP_RTI);
	
	jtag_add_dr_scan(3, fields, -1);
	
	return ERROR_OK;
}

int xscale_receive(target_t *target, u32 *buffer, int num_words)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	struct timeval timeout, now;
	
	enum tap_state path[3];
	scan_field_t fields[3];
	
	u8 *field0 = malloc(num_words * 1);
	u32 *field1 = malloc(num_words * 4);
	int words_done = 0;
	int words_scheduled = 0;
	
	int i;
	int retval;
	
	path[0] = TAP_SDS;
	path[1] = TAP_CD;
	path[2] = TAP_SD;
	
	fields[0].device = xscale->chain_pos;
	fields[0].num_bits = 3;
	fields[0].out_value = NULL;
	fields[0].out_mask = NULL;
	/* fields[0].in_value = field0; */
	fields[0].in_check_value = NULL;
	fields[0].in_check_mask = NULL;
	fields[0].in_handler = NULL;
	fields[0].in_handler_priv = NULL;
		
	fields[1].device = xscale->chain_pos;
	fields[1].num_bits = 32;
	fields[1].out_value = NULL;
	fields[1].out_mask = NULL;
	/* fields[1].in_value = field1; */
	fields[1].in_handler = NULL;
	fields[1].in_handler_priv = NULL;
	fields[1].in_check_value = NULL;
	fields[1].in_check_mask = NULL;

	fields[2].device = xscale->chain_pos;
	fields[2].num_bits = 1;
	fields[2].out_value = NULL;
	fields[2].out_mask = NULL;
	fields[2].in_value = NULL;
	fields[2].in_check_value = NULL;
	fields[2].in_check_mask = NULL;
	fields[2].in_handler = NULL;
	fields[2].in_handler_priv = NULL;

	jtag_add_end_state(TAP_RTI);
	xscale_jtag_set_instr(xscale->chain_pos, 0x10); /* DBGTX */
	
	/* repeat until all words have been collected */
	while (words_done < num_words)
	{
		/* schedule reads */
		words_scheduled = 0;
		for (i = words_done; i < num_words; i++)
		{
			fields[0].in_value = &field0[i];
			fields[1].in_value = (u8*)&field1[i];
			
			jtag_add_pathmove(3, path);
			jtag_add_dr_scan(3, fields, TAP_RTI);
			words_scheduled++;
		}
		
		if ((retval = jtag_execute_queue()) != ERROR_OK)
		{
			ERROR("JTAG error while reading TX");
			exit(-1);
		}
		
		/* examine results */
		for (i = words_done; i < num_words; i++)
		{
			if (!(field0[0] & 1))
			{
				/* move backwards if necessary */
				int j;
				for (j = i; j < num_words - 1; j++)
				{
					field0[j] = field0[j+1];
					field1[j] = field1[j+1];
				}
				words_scheduled--;
			}
		}
		words_done += words_scheduled;
	}
	
	for (i = 0; i < num_words; i++)
		*(buffer++) = buf_get_u32((u8*)&field1[i], 0, 32);
	
	free(field0);
	free(field1);
	
	return ERROR_OK;
}

int xscale_read_tx(target_t *target, int consume)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	enum tap_state path[3];
	
	int retval;
	struct timeval timeout, now;
	
	scan_field_t fields[3];
	u8 field0 = 0x0;
	u8 field2 = 0x0;

	jtag_add_end_state(TAP_RTI);
	
	xscale_jtag_set_instr(xscale->chain_pos, 0x10); /* DBGTX */
	
	path[0] = TAP_SDS;
	path[1] = TAP_CD;
	path[2] = TAP_SD;
	
	fields[0].device = xscale->chain_pos;
	fields[0].num_bits = 3;
	fields[0].out_value = NULL;
	fields[0].out_mask = NULL;
	fields[0].in_value = &field0;
	fields[0].in_check_value = NULL;
	fields[0].in_check_mask = NULL;
	fields[0].in_handler = NULL;
	fields[0].in_handler_priv = NULL;
		
	fields[1].device = xscale->chain_pos;
	fields[1].num_bits = 32;
	fields[1].out_value = NULL;
	fields[1].out_mask = NULL;
	fields[1].in_value = xscale->reg_cache->reg_list[XSCALE_TX].value;
	fields[1].in_handler = NULL;
	fields[1].in_handler_priv = NULL;
	fields[1].in_check_value = NULL;
	fields[1].in_check_mask = NULL;

	fields[2].device = xscale->chain_pos;
	fields[2].num_bits = 1;
	fields[2].out_value = NULL;
	fields[2].out_mask = NULL;
	fields[2].in_value = &field2;
	fields[2].in_check_value = NULL;
	fields[2].in_check_mask = NULL;
	fields[2].in_handler = NULL;
	fields[2].in_handler_priv = NULL;
	
	gettimeofday(&timeout, NULL);
	timeval_add_time(&timeout, 5, 0);
	
	do
	{
		/* if we want to consume the register content (i.e. clear TX_READY),
		 * we have to go straight from Capture-DR to Shift-DR
		 * otherwise, we go from Capture-DR to Exit1-DR to Pause-DR
		*/
		if (consume)
			jtag_add_pathmove(3, path);
		else
			jtag_add_statemove(TAP_PD);
		
		jtag_add_dr_scan(3, fields, TAP_RTI);
	
		if ((retval = jtag_execute_queue()) != ERROR_OK)
		{
			ERROR("JTAG error while reading DCSR");
			exit(-1);
		}
		
		gettimeofday(&now, NULL);
		if ((now.tv_sec > timeout.tv_sec) && (now.tv_usec > timeout.tv_usec))
		{
			ERROR("time out reading TX register");
			return ERROR_TARGET_TIMEOUT;
		}
	} while ((!(field0 & 1)) && consume);
	
	if (!(field0 & 1))
		return ERROR_TARGET_RESOURCE_NOT_AVAILABLE;
		
	return ERROR_OK;
}

int xscale_write_rx(target_t *target)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	int retval;
	struct timeval timeout, now;
	
	scan_field_t fields[3];
	u8 field0_out = 0x0;
	u8 field0_in = 0x0;
	u8 field2 = 0x0;

	jtag_add_end_state(TAP_RTI);
	
	xscale_jtag_set_instr(xscale->chain_pos, 0x02); /* DBGRX */
	
	fields[0].device = xscale->chain_pos;
	fields[0].num_bits = 3;
	fields[0].out_value = &field0_out;
	fields[0].out_mask = NULL;
	fields[0].in_value = &field0_in;
	fields[0].in_check_value = NULL;
	fields[0].in_check_mask = NULL;
	fields[0].in_handler = NULL;
	fields[0].in_handler_priv = NULL;
		
	fields[1].device = xscale->chain_pos;
	fields[1].num_bits = 32;
	fields[1].out_value = xscale->reg_cache->reg_list[XSCALE_RX].value;
	fields[1].out_mask = NULL;
	fields[1].in_value = NULL;
	fields[1].in_handler = NULL;
	fields[1].in_handler_priv = NULL;
	fields[1].in_check_value = NULL;
	fields[1].in_check_mask = NULL;

	fields[2].device = xscale->chain_pos;
	fields[2].num_bits = 1;
	fields[2].out_value = &field2;
	fields[2].out_mask = NULL;
	fields[2].in_value = NULL;
	fields[2].in_check_value = NULL;
	fields[2].in_check_mask = NULL;
	fields[2].in_handler = NULL;
	fields[2].in_handler_priv = NULL;
	
	gettimeofday(&timeout, NULL);
	timeval_add_time(&timeout, 5, 0);
	
	/* poll until rx_read is low */
	do
	{
		DEBUG("polling RX");
		jtag_add_dr_scan(3, fields, TAP_RTI);
	
		if ((retval = jtag_execute_queue()) != ERROR_OK)
		{
			ERROR("JTAG error while reading DCSR");
			exit(-1);
		}
		
		gettimeofday(&now, NULL);
		if ((now.tv_sec > timeout.tv_sec) && (now.tv_usec > timeout.tv_usec))
		{
			ERROR("time out reading TX register");
			return ERROR_TARGET_TIMEOUT;
		}
	} while (field0_in & 1);
	
	/* set rx_valid */
	field2 = 0x1;
	jtag_add_dr_scan(3, fields, TAP_RTI);
	
	if ((retval = jtag_execute_queue()) != ERROR_OK)
	{
		ERROR("JTAG error while reading DCSR");
		exit(-1);
	}
	
	return ERROR_OK;
}

int xscale_write_dcsr(target_t *target, int hold_rst, int ext_dbg_brk)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	int retval;
	
	scan_field_t fields[3];
	u8 field0 = 0x0;
	u8 field2 = 0x0;
	
	if (hold_rst != -1)
		xscale->hold_rst = hold_rst;
	
	if (ext_dbg_brk != -1)
		xscale->external_debug_break = ext_dbg_brk;

	jtag_add_end_state(TAP_RTI);
	xscale_jtag_set_instr(xscale->chain_pos, 0x09); /* DCSR */
	
	buf_set_u32(&field0, 1, 1, xscale->hold_rst);
	buf_set_u32(&field0, 2, 1, xscale->external_debug_break);
	
	fields[0].device = xscale->chain_pos;
	fields[0].num_bits = 3;
	fields[0].out_value = &field0;
	fields[0].out_mask = NULL;
	fields[0].in_value = NULL;
	fields[0].in_check_value = NULL;
	fields[0].in_check_mask = NULL;
	fields[0].in_handler = NULL;
	fields[0].in_handler_priv = NULL;
		
	fields[1].device = xscale->chain_pos;
	fields[1].num_bits = 32;
	fields[1].out_value = xscale->reg_cache->reg_list[XSCALE_DCSR].value;
	fields[1].out_mask = NULL;
	fields[1].in_value = NULL;
	fields[1].in_handler = NULL;
	fields[1].in_handler_priv = NULL;
	fields[1].in_check_value = NULL;
	fields[1].in_check_mask = NULL;

	fields[2].device = xscale->chain_pos;
	fields[2].num_bits = 1;
	fields[2].out_value = &field2;
	fields[2].out_mask = NULL;
	fields[2].in_value = NULL;
	fields[2].in_check_value = NULL;
	fields[2].in_check_mask = NULL;
	fields[2].in_handler = NULL;
	fields[2].in_handler_priv = NULL;
	
	jtag_add_dr_scan(3, fields, -1);
	
	if ((retval = jtag_execute_queue()) != ERROR_OK)
	{
		ERROR("JTAG error while writing DCSR");
		exit(-1);
	}
	
	xscale->reg_cache->reg_list[XSCALE_DCSR].dirty = 0;
	xscale->reg_cache->reg_list[XSCALE_DCSR].valid = 1;
	
	return ERROR_OK;
}

/* parity of the number of bits 0 if even; 1 if odd. for 32 bit words */
unsigned int parity (unsigned int v)
{
	unsigned int ov = v;
	v ^= v >> 16;
	v ^= v >> 8;
	v ^= v >> 4;
	v &= 0xf;
	DEBUG("parity of 0x%x is %i", ov, (0x6996 >> v) & 1);
	return (0x6996 >> v) & 1;
}

int xscale_load_ic(target_t *target, int mini, u32 va, u8 *buffer)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	u8 packet[4];
	u8 cmd;
	int word, byte;
	
	scan_field_t fields[2];

	jtag_add_end_state(TAP_RTI);
	xscale_jtag_set_instr(xscale->chain_pos, 0x07); /* LDIC */
	
	/* CMD is b010 for Main IC and b011 for Mini IC */
	if (mini)
		buf_set_u32(&cmd, 0, 3, 0x3);
	else
		buf_set_u32(&cmd, 0, 3, 0x2);
	
	buf_set_u32(&cmd, 3, 3, 0x0);
		
	/* virtual address of desired cache line */
	buf_set_u32(packet, 0, 27, va >> 5);
	
	fields[0].device = xscale->chain_pos;
	fields[0].num_bits = 6;
	fields[0].out_value = &cmd;
	fields[0].out_mask = NULL;
	fields[0].in_value = NULL;
	fields[0].in_check_value = NULL;
	fields[0].in_check_mask = NULL;
	fields[0].in_handler = NULL;
	fields[0].in_handler_priv = NULL;

	fields[1].device = xscale->chain_pos;
	fields[1].num_bits = 27;
	fields[1].out_value = packet;
	fields[1].out_mask = NULL;
	fields[1].in_value = NULL;
	fields[1].in_check_value = NULL;
	fields[1].in_check_mask = NULL;
	fields[1].in_handler = NULL;
	fields[1].in_handler_priv = NULL;
	
	jtag_add_dr_scan(2, fields, -1);

	fields[0].num_bits = 32;
	fields[0].out_value = packet;
	
	fields[1].num_bits = 1;
	fields[1].out_value = &cmd;
	
	for (word = 0; word < 8; word++)
	{
		for (byte = 0; byte < 4; byte++)
			packet[byte] = buffer[byte + 4*word];
		cmd = parity(*((u32*)packet));
		jtag_add_dr_scan(2, fields, -1);
	}
	
	jtag_execute_queue();
	
	return ERROR_OK;
}

enum target_state xscale_poll(target_t *target)
{
	int retval;
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	if ((target->state == TARGET_RUNNING) || (target->state == TARGET_DEBUG_RUNNING))
	{
		if ((retval = xscale_read_tx(target, 0)) == ERROR_OK)
		{
			/* there's data to read from the tx register, we entered debug state */
			target->state = TARGET_HALTED;
			
			/* process debug entry, fetching current mode regs */
			if ((retval = xscale_debug_entry(target)) != ERROR_OK)
				return retval;
			
			/* if target was running, signal that we halted
			 * otherwise we reentered from debug execution */
			if (target->state == TARGET_RUNNING)
				target_call_event_callbacks(target, TARGET_EVENT_HALTED);
			else
				target_call_event_callbacks(target, TARGET_EVENT_DEBUG_HALTED);
		}
		else if (retval != ERROR_TARGET_RESOURCE_NOT_AVAILABLE)
		{
			ERROR("error while polling TX register");
			exit(-1);
		}
	}
	
	return target->state;
}

int xscale_debug_entry(target_t *target)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	u32 *buffer = malloc(4 * 10);
	int i;
	
	u32 moe;
	
	/* get r0 to r7, pc and cpsr */
	xscale_receive(target, buffer, 10);
	
	/* move data from buffer to register cache */
	for (i = 0; i <= 7; i++)
	{
		buf_set_u32(armv4_5->core_cache->reg_list[i].value, 0, 32, buf_get_u32(buffer + i, 0, 32));
		armv4_5->core_cache->reg_list[i].dirty = 1;
		armv4_5->core_cache->reg_list[i].valid = 1;
	}
	
	buf_set_u32(armv4_5->core_cache->reg_list[15].value, 0, 32, buf_get_u32(buffer + 8, 0, 32));
	armv4_5->core_cache->reg_list[15].dirty = 1;
	armv4_5->core_cache->reg_list[15].valid = 1;
		
	buf_set_u32(armv4_5->core_cache->reg_list[ARMV4_5_CPSR].value, 0, 32, buf_get_u32(buffer + 9, 0, 32));
	armv4_5->core_cache->reg_list[ARMV4_5_CPSR].dirty = 1;
	armv4_5->core_cache->reg_list[ARMV4_5_CPSR].valid = 1;
	
	armv4_5->core_mode = buf_get_u32(buffer + 9, 0, 32) & 0x1f;
	if (armv4_5_mode_to_number(armv4_5->core_mode) == -1)
	{
		target->state = TARGET_UNKNOWN;
		ERROR("cpsr contains invalid mode value - communication failure");
		return ERROR_TARGET_FAILURE;
	}
	DEBUG("target entered debug state in %s mode", armv4_5_mode_strings[armv4_5_mode_to_number(armv4_5->core_mode)]);
	
	if (buf_get_u32(buffer + 9, 0, 32) & 0x20)
		armv4_5->core_state = ARMV4_5_STATE_THUMB;
	else
		armv4_5->core_state = ARMV4_5_STATE_ARM;
	
	/* get banked registers, r8 to r14, and spsr if not in USR/SYS mode */
	if ((armv4_5->core_mode != ARMV4_5_MODE_USR) && (armv4_5->core_mode != ARMV4_5_MODE_SYS))
	{
		xscale_receive(target, buffer, 8);
		buf_set_u32(ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5->core_mode, 16).value, 0, 32, buf_get_u32(buffer + 7, 0, 32));
		ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5->core_mode, 16).dirty = 0;
		ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5->core_mode, 16).valid = 1;
	}
	else
	{
		xscale_receive(target, buffer, 7);
	}
	
	/* move data from buffer to register cache */
	for (i = 8; i <= 14; i++)
	{
		buf_set_u32(ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5->core_mode, i).value, 0, 32, buf_get_u32(buffer + (i - 8), 0, 32));
		ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5->core_mode, i).dirty = 0;
		ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5->core_mode, i).valid = 1;
	}
	
	/* examine debug reason */
	xscale_read_dcsr(target);
	moe = buf_get_u32(xscale->reg_cache->reg_list[XSCALE_DCSR].value, 2, 3);
	
	switch (moe)
	{
		case 0x0: /* Processor reset */
			buf_set_u32(armv4_5->core_cache->reg_list[15].value, 0, 32, 0x0);
			target->debug_reason = DBG_REASON_DBGRQ;
			break;
		case 0x1: /* Instruction breakpoint hit */
			target->debug_reason = DBG_REASON_BREAKPOINT;
			break;
		case 0x2: /* Data breakpoint hit */
			target->debug_reason = DBG_REASON_WATCHPOINT;
			break;
		case 0x3: /* BKPT instruction executed */
			target->debug_reason = DBG_REASON_BREAKPOINT;
			break;
		case 0x4: /* Ext. debug event */
			target->debug_reason = DBG_REASON_DBGRQ;
			break;
		case 0x5: /* Vector trap occured */
			target->debug_reason = DBG_REASON_BREAKPOINT;
			break;
		case 0x6: /* Trace buffer full break */
			target->debug_reason = DBG_REASON_DBGRQ;
			break;
		case 0x7: /* Reserved */
		default:
			ERROR("Method of Entry is 'Reserved'");
			exit(-1);
			break;
	}
	
	return ERROR_OK;
}

int xscale_halt(target_t *target)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	DEBUG("target->state: %s", target_state_strings[target->state]);
	
	if (target->state == TARGET_HALTED)
	{
		WARNING("target was already halted");
		return ERROR_TARGET_ALREADY_HALTED;
	} 
	else if (target->state == TARGET_UNKNOWN)
	{
		/* this must not happen for a xscale target */
		ERROR("target was in unknown state when halt was requested");
		exit(-1);
	}
	else if (target->state == TARGET_RESET)
	{
		DEBUG("target->state == TARGET_RESET");
	}

	/* TODO: signal debug request */

	target->debug_reason = DBG_REASON_DBGRQ;
	
	return ERROR_OK;
}

int xscale_resume(struct target_s *target, int current, u32 address, int handle_breakpoints, int debug_execution)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale= armv4_5->arch_info;
	breakpoint_t *breakpoint = target->breakpoints;
	
	u32 current_pc;
	
	DEBUG("");
	
	if (target->state != TARGET_HALTED)
	{
		WARNING("target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}
	
	if (!debug_execution)
	{
		target_free_all_working_areas(target);
	}
	
	/* current = 1: continue on current pc, otherwise continue at <address> */
	if (!current)
		buf_set_u32(armv4_5->core_cache->reg_list[15].value, 0, 32, address);
	else
		current_pc = buf_get_u32(armv4_5->core_cache->reg_list[15].value, 0, 32);
	
	/* the front-end may request us not to handle breakpoints */
	if (handle_breakpoints)
	{
		if ((breakpoint = breakpoint_find(target, buf_get_u32(armv4_5->core_cache->reg_list[15].value, 0, 32))))
		{
			
		}
	}
	
	/* enable any pending breakpoints and watchpoints */
	//xscale_enable_breakpoints(target);
	//xscale_enable_watchpoints(target);
	
	/* restore banked registers */
	xscale_restore_context(target);
	
	/* if we're at the reset vector, we have to simulate the branch */
	if (current_pc == 0x0)
		current_pc = armv4_5_next_pc(target);
		
	target->debug_reason = DBG_REASON_NOTHALTED;
	
	if (!debug_execution)
	{
		/* registers are now invalid */
		armv4_5_invalidate_core_regs(target);
		target->state = TARGET_RUNNING;
		target_call_event_callbacks(target, TARGET_EVENT_RESUMED);
	}
	else
	{
		target->state = TARGET_DEBUG_RUNNING;
		target_call_event_callbacks(target, TARGET_EVENT_DEBUG_RESUMED);
	}
	
	DEBUG("target resumed");
	
	return ERROR_OK;
}

int xscale_step(struct target_s *target, int current, u32 address, int handle_breakpoints)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	breakpoint_t *breakpoint = target->breakpoints;

	if (target->state != TARGET_HALTED)
	{
		WARNING("target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}
	
	/* current = 1: continue on current pc, otherwise continue at <address> */
	if (!current)
		buf_set_u32(armv4_5->core_cache->reg_list[15].value, 0, 32, address);
	
	/* the front-end may request us not to handle breakpoints */
	if (handle_breakpoints)
		if ((breakpoint = breakpoint_find(target, buf_get_u32(armv4_5->core_cache->reg_list[15].value, 0, 32))))
		{
		}
	
	target->debug_reason = DBG_REASON_SINGLESTEP;

	target_call_event_callbacks(target, TARGET_EVENT_RESUMED);

	/* registers are now invalid */
	armv4_5_invalidate_core_regs(target);
	
	target_call_event_callbacks(target, TARGET_EVENT_HALTED);

	if (breakpoint)
	{
	}
		
	DEBUG("target stepped");

	return ERROR_OK;

}

int xscale_assert_reset(target_t *target)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	int retval;
	
	DEBUG("target->state: %s", target_state_strings[target->state]);
	
	/* if the handler isn't installed yet, we have to assert TRST, too */
	if (!xscale->handler_installed)
	{
		jtag_add_reset(1, 1);
	}
	else
		jtag_add_reset(-1, 1);
	
	/* sleep 1ms, to be sure we fulfill any requirements */
	jtag_add_sleep(1000);
	
	target->state = TARGET_RESET;
	
	return ERROR_OK;

}

int xscale_deassert_reset(target_t *target)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	FILE *binary;
	u32 address;
	struct stat binary_stat;
	u32 binary_size;

	u8 *buffer;
	u32 buf_cnt;
	
	if (!xscale->handler_installed)
	{
		/* release TRST */
		jtag_add_reset(0, -1);
		jtag_add_sleep(100000);
		
		/* set Hold reset, Halt mode and Trap Reset */
		buf_set_u32(xscale->reg_cache->reg_list[XSCALE_DCSR].value, 30, 1, 0x1);
		buf_set_u32(xscale->reg_cache->reg_list[XSCALE_DCSR].value, 16, 1, 0x1);
		xscale_write_dcsr(target, 1, 0);
		jtag_add_runtest(100, TAP_RTI);
		jtag_execute_queue();
		
		/* release SRST */
		jtag_add_reset(0, 0);
		/* wait 150ms; 100ms were not enough */
		jtag_add_sleep(150000);

		jtag_add_runtest(2030, TAP_RTI);
		jtag_execute_queue();
		
		xscale_write_dcsr(target, 1, 0);
		jtag_execute_queue();
		
		/* TODO: load debug handler */
		if (stat("target/xscale/debug_handler.bin", &binary_stat) == -1)
		{
			ERROR("couldn't stat() target/xscale/debug_handler.bin: %s",  strerror(errno));
			return ERROR_OK;
		}
		
		if (!(binary = fopen("target/xscale/debug_handler.bin", "r")))
		{
			ERROR("couldn't open target/xscale/debug_handler.bin: %s", strerror(errno));
			return ERROR_OK;
		}
		
		buffer = malloc(32);
		
		if ((binary_size = binary_stat.st_size) % 4)
			ERROR("debug_handler.bin: size not a multiple of 4");
		binary_size = CEIL(binary_size, 32) * 32;
		
		address = 0x0;
		while (binary_size > 0)
		{
			buf_cnt = fread(buffer, 1, 8*4, binary);
			if (buf_cnt < 8*4)
			{
				for (; buf_cnt < 8*4; buf_cnt += 4)
				{
					u32 *pad = (u32*)buffer + buf_cnt;
					*pad = 0xe1a08008;
				}
			}
			xscale_load_ic(target, 1, address, buffer);
			address += buf_cnt;
			binary_size -= buf_cnt;
		};
		
		free(buffer);
		
		jtag_add_runtest(30, TAP_RTI);
		
		/* let the target run (should enter debug handler) */
		xscale_write_dcsr(target, 0, 0);
		target->state = TARGET_RUNNING;
		
		if ((target->reset_mode != RESET_HALT) && (target->reset_mode != RESET_INIT))
		{
			jtag_add_sleep(10000);
			
			/* we should have entered debug now */
			xscale_debug_entry(target);
			target->state = TARGET_HALTED;
			
			/* the PC is now at 0x0 */
			buf_set_u32(armv4_5->core_cache->reg_list[15].value, 0, 32, 0x0);
		}
	}
	else
	{
		jtag_add_reset(0, 0);
	}
		
		
	return ERROR_OK;
}

int xscale_soft_reset_halt(struct target_s *target)
{
	
	return ERROR_OK;
}

int xscale_read_core_reg(struct target_s *target, int num, enum armv4_5_mode mode)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	return ERROR_OK;
}

int xscale_write_core_reg(struct target_s *target, int num, enum armv4_5_mode mode, u32 value)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	return ERROR_OK;
}

int xscale_full_context(target_t *target)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	u32 *buffer;
	
	int i, j;
	
	DEBUG("");
	
	if (target->state != TARGET_HALTED)
	{
		WARNING("target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}
	
	buffer = malloc(4 * 8);
	
	/* iterate through processor modes (FIQ, IRQ, SVC, ABT, UND and SYS)
	 * we can't enter User mode on an XScale (unpredictable),
	 * but User shares registers with SYS
	 */
	for(i = 1; i < 7; i++)
	{
		int valid = 1;
		
		/* check if there are invalid registers in the current mode 
		 */
		for (j = 0; j <= 16; j++)
		{
			if (ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), j).valid == 0)
				valid = 0;
		}
		
		if (!valid)
		{
			u32 tmp_cpsr;
			
			/* request banked registers */
			buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, 0x0);
			xscale_write_rx(target);
			
			tmp_cpsr = 0x0;
			tmp_cpsr |= armv4_5_number_to_mode(i);
			tmp_cpsr |= 0xc0; /* I/F bits */
			
			/* send CPSR for desired mode */
			buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, tmp_cpsr);
			xscale_write_rx(target);

			/* get banked registers, r8 to r14, and spsr if not in USR/SYS mode */
			if ((armv4_5_number_to_mode(i) != ARMV4_5_MODE_USR) && (armv4_5_number_to_mode(i) != ARMV4_5_MODE_SYS))
			{
				xscale_receive(target, buffer, 8);
				buf_set_u32(ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5->core_mode, 16).value, 0, 32, buf_get_u32(buffer + 7, 0, 32));
				ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), 16).dirty = 0;
				ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), 16).valid = 1;
			}
			else
			{
				xscale_receive(target, buffer, 7);
			}
	
			/* move data from buffer to register cache */
			for (j = 8; j <= 14; j++)
			{
				buf_set_u32(ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), j).value, 0, 32, buf_get_u32(buffer + (j - 8), 0, 32));
				ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), j).dirty = 0;
				ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), j).valid = 1;
			}
		}
	}
	
	free(buffer);
	
	return ERROR_OK;
}

int xscale_restore_context(target_t *target)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	int i, j;
	
	DEBUG("");
	
	if (target->state != TARGET_HALTED)
	{
		WARNING("target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}
	
	/* iterate through processor modes (FIQ, IRQ, SVC, ABT, UND and SYS)
	* we can't enter User mode on an XScale (unpredictable),
	* but User shares registers with SYS
	*/
	for(i = 1; i < 7; i++)
	{
		int dirty = 0;
		
		/* check if there are invalid registers in the current mode 
		*/
		for (j = 8; j <= 14; j++)
		{
			if (ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), j).dirty == 1)
				dirty = 1;
		}
		
		/* if not USR/SYS, check if the SPSR needs to be written */
		if ((armv4_5_number_to_mode(i) != ARMV4_5_MODE_USR) && (armv4_5_number_to_mode(i) != ARMV4_5_MODE_SYS))
		{
			if (ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), 16).dirty == 1)
				dirty = 1;
		}
		
		if (dirty)
		{
			u32 tmp_cpsr;
			
			/* send banked registers */
			buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, 0x1);
			xscale_write_rx(target);
			
			tmp_cpsr = 0x0;
			tmp_cpsr |= armv4_5_number_to_mode(i);
			tmp_cpsr |= 0xc0; /* I/F bits */
			
			/* send CPSR for desired mode */
			buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, tmp_cpsr);
			xscale_write_rx(target);

			/* send banked registers, r8 to r14, and spsr if not in USR/SYS mode */
			for (j = 8; j <= 14; j++)
			{
				buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, buf_get_u32(ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5->core_mode, j).value, 0, 32));
				ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), j).dirty = 0;
				xscale_write_rx(target);
			}
			
			if ((armv4_5_number_to_mode(i) != ARMV4_5_MODE_USR) && (armv4_5_number_to_mode(i) != ARMV4_5_MODE_SYS))
			{
				buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, buf_get_u32(ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5->core_mode, 16).value, 0, 32));
				ARMV4_5_CORE_REG_MODE(armv4_5->core_cache, armv4_5_number_to_mode(i), 16).dirty = 0;
				xscale_write_rx(target);
			}
	

		}
	}
	
	return ERROR_OK;
}

int xscale_read_memory(struct target_s *target, u32 address, u32 size, u32 count, u8 *buffer)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	u32 *buf32;
	int i;
	
	DEBUG("address: 0x%8.8x, size: 0x%8.8x, count: 0x%8.8x", address, size, count);

	if (target->state != TARGET_HALTED)
	{
		WARNING("target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}

	/* sanitize arguments */
	if (((size != 4) && (size != 2) && (size != 1)) || (count == 0) || !(buffer))
		return ERROR_INVALID_ARGUMENTS;

	if (((size == 4) && (address & 0x3u)) || ((size == 2) && (address & 0x1u)))
		return ERROR_TARGET_UNALIGNED_ACCESS;
	
	/* send memory read request (command 0x1n, n: access size) */
	buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, 0x10 | size);
	xscale_write_rx(target);
	
	/* send base address for read request */
	buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, address);
	xscale_write_rx(target);
	
	/* send number of requested data words */
	buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, count);
	xscale_write_rx(target);
	
	/* receive data from target (count times 32-bit words in host endianness) */
	buf32 = malloc(4 * count);
	xscale_receive(target, buf32, count);
	
	/* extract data from host-endian buffer into byte stream */
	for (i = 0; i < count; i++)
	{
		switch (size)
		{
			case 4:
				target_buffer_set_u32(target, buffer, buf32[i]);
				buffer += 4;
				break;
			case 2:
				target_buffer_set_u16(target, buffer, buf32[i] & 0xffff);
				buffer += 2;
				break;
			case 1:
				*buffer++ = buf32[i] & 0xff;
				break;
			default:
				ERROR("should never get here");
				exit(-1);
		}
	}
	
	free(buf32);
	
	return ERROR_OK;
}

int xscale_write_memory(struct target_s *target, u32 address, u32 size, u32 count, u8 *buffer)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	DEBUG("address: 0x%8.8x, size: 0x%8.8x, count: 0x%8.8x", address, size, count);

	if (target->state != TARGET_HALTED)
	{
		WARNING("target not halted");
		return ERROR_TARGET_NOT_HALTED;
	}

	/* sanitize arguments */
	if (((size != 4) && (size != 2) && (size != 1)) || (count == 0) || !(buffer))
		return ERROR_INVALID_ARGUMENTS;

	if (((size == 4) && (address & 0x3u)) || ((size == 2) && (address & 0x1u)))
		return ERROR_TARGET_UNALIGNED_ACCESS;
	
	/* send memory read request (command 0x1n, n: access size) */
	buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, 0x10 | size);
	xscale_write_rx(target);
	
	/* send base address for read request */
	buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, address);
	xscale_write_rx(target);
	
	/* send number of requested data words */
	buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, count);
	xscale_write_rx(target);
	
	return ERROR_OK;
}

int xscale_bulk_write_memory(target_t *target, u32 address, u32 count, u8 *buffer)
{
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	xscale_write_memory(target, address, 4, count, buffer);
	
	return ERROR_OK;
}

int xscale_add_breakpoint(struct target_s *target, u32 address, u32 length, enum breakpoint_type type)
{
	
}

int xscale_remove_breakpoint(struct target_s *target, breakpoint_t *breakpoint)
{
	
}

int xscale_add_watchpoint(struct target_s *target, u32 address, u32 length, enum watchpoint_rw rw)
{
	
}

int xscale_remove_watchpoint(struct target_s *target, watchpoint_t *watchpoint)
{
	
}

int xscale_get_reg(reg_t *reg)
{
	xscale_reg_t *arch_info = reg->arch_info;
	target_t *target = arch_info->target;
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
		
	/* DCSR, TX and RX are accessible via JTAG */
	if (strcmp(reg->name, "XSCALE_DCSR") == 0)
	{
		return xscale_read_dcsr(arch_info->target);
	}
	else if (strcmp(reg->name, "XSCALE_TX") == 0)
	{
		/* 1 = consume register content */
		return xscale_read_tx(arch_info->target, 1);
	}
	else if (strcmp(reg->name, "XSCALE_RX") == 0)
	{
		/* can't read from RX register (host -> debug handler) */
		return ERROR_OK;
	}
	else if (strcmp(reg->name, "XSCALE_TXRXCTRL") == 0)
	{
		/* can't (explicitly) read from TXRXCTRL register */
		return ERROR_OK;
	}
	else /* Other DBG registers have to be transfered by the debug handler */
	{
		/* send CP read request (command 0x40) */
		buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, 0x40);
		xscale_write_rx(target);
		
		/* send CP read request (command 0x40) */
		buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, arch_info->dbg_handler_number);
		xscale_write_rx(target);
		
		xscale_receive(target, reg->value, 1);
		reg->dirty = 0;
		reg->valid = 1;
	}
	
	return ERROR_OK;
}

int xscale_set_reg(reg_t *reg, u32 value)
{
	xscale_reg_t *arch_info = reg->arch_info;
	target_t *target = arch_info->target;
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;
	
	/* DCSR, TX and RX are accessible via JTAG */
	if (strcmp(reg->name, "XSCALE_DCSR") == 0)
	{
		buf_set_u32(xscale->reg_cache->reg_list[XSCALE_DCSR].value, 0, 32, value);
		return xscale_write_dcsr(arch_info->target, -1, -1);
	}
	else if (strcmp(reg->name, "XSCALE_RX") == 0)
	{
		buf_set_u32(xscale->reg_cache->reg_list[XSCALE_RX].value, 0, 32, value);
		return xscale_write_rx(arch_info->target);
	}
	else if (strcmp(reg->name, "XSCALE_TX") == 0)
	{
		/* can't write to TX register (debug-handler -> host) */
		return ERROR_OK;
	}
	else if (strcmp(reg->name, "XSCALE_TXRXCTRL") == 0)
	{
		/* can't (explicitly) write to TXRXCTRL register */
		return ERROR_OK;
	}
	else /* Other DBG registers have to be transfered by the debug handler */
	{
		
	}
}

void xscale_build_reg_cache(target_t *target)
{
	/* get pointers to arch-specific information */
	armv4_5_common_t *armv4_5 = target->arch_info;
	xscale_common_t *xscale = armv4_5->arch_info;

	reg_cache_t **cache_p = register_get_last_cache_p(&target->reg_cache);
	xscale_reg_t *arch_info = malloc(sizeof(xscale_reg_arch_info));
	int i;
	int num_regs = sizeof(xscale_reg_arch_info) / sizeof(xscale_reg_t);
	
	(*cache_p) = armv4_5_build_reg_cache(target, armv4_5);
	armv4_5->core_cache = (*cache_p);
	
	/* register a register arch-type for XScale dbg registers only once */
	if (xscale_reg_arch_type == -1)
		xscale_reg_arch_type = register_reg_arch_type(xscale_get_reg, xscale_set_reg);
	
	(*cache_p)->next = malloc(sizeof(reg_cache_t));
	cache_p = &(*cache_p)->next;
	
	/* fill in values for the xscale reg cache */
	(*cache_p)->name = "XScale registers";
	(*cache_p)->next = NULL;
	(*cache_p)->reg_list = malloc(num_regs * sizeof(reg_t));
	(*cache_p)->num_regs = num_regs;
	
	for (i = 0; i < num_regs; i++)
	{
		(*cache_p)->reg_list[i].name = xscale_reg_list[i];
		(*cache_p)->reg_list[i].value = calloc(4, 1);
		(*cache_p)->reg_list[i].dirty = 0;
		(*cache_p)->reg_list[i].valid = 0;
		(*cache_p)->reg_list[i].size = 32;
		(*cache_p)->reg_list[i].bitfield_desc = NULL;
		(*cache_p)->reg_list[i].num_bitfields = 0;
		(*cache_p)->reg_list[i].arch_info = &arch_info[i];
		(*cache_p)->reg_list[i].arch_type = xscale_reg_arch_type;
		arch_info[i] = xscale_reg_arch_info[i];
		arch_info[i].target = target;
	}
	
	xscale->reg_cache = (*cache_p);
}

int xscale_init_target(struct command_context_s *cmd_ctx, struct target_s *target)
{
	if (startup_mode != DAEMON_RESET)
	{
		ERROR("XScale target requires a reset");
		ERROR("Reset target to enable debug");
	}
	
	return ERROR_OK;
}

int xscale_quit()
{
	
	return ERROR_OK;
}

int xscale_init_arch_info(target_t *target, xscale_common_t *xscale, int chain_pos, char *variant)
{
	armv4_5_common_t *armv4_5;
	
	armv4_5 = &xscale->armv4_5_common;
	
	/* prepare JTAG information for the new target */
	xscale->chain_pos = chain_pos;
	jtag_register_event_callback(xscale_jtag_callback, target);
	
	xscale->arch_info = NULL;
	xscale->common_magic = XSCALE_COMMON_MAGIC;
	
	xscale->variant = strdup(variant);
	
	xscale->handler_installed = 0;
	
	armv4_5->arch_info = xscale;
	armv4_5->read_core_reg = xscale_read_core_reg;
	armv4_5->write_core_reg = xscale_write_core_reg;
	armv4_5->full_context = xscale_full_context;
	
	armv4_5_init_arch_info(target, armv4_5);
	
	return ERROR_OK;
}

/* target xscale <endianess> <startup_mode> <chain_pos> <variant> */
int xscale_target_command(struct command_context_s *cmd_ctx, char *cmd, char **args, int argc, struct target_s *target)
{
	int chain_pos;
	char *variant = NULL;
	xscale_common_t *xscale = malloc(sizeof(xscale_common_t));

	if (argc < 5)
	{
		ERROR("'target xscale' requires at least one additional argument");
		exit(-1);
	}
	
	chain_pos = strtoul(args[2], NULL, 0);
	
	variant = args[4];
	
	xscale_init_arch_info(target, xscale, chain_pos, variant);
	xscale_build_reg_cache(target);
	
	return ERROR_OK;
}

int xscale_register_commands(struct command_context_s *cmd_ctx)
{
	
	armv4_5_register_commands(cmd_ctx);
	
	return ERROR_OK;
}